(* Alpha-conversion *)

(*s This module performs an alpha-conversion pass of programs. 
  We freshly rename all binders and data constructors. *)

open Ast
open Utils

(* The module Env maps old identifiers to fresh ones. *)

type strdata = {
  idents : (string * string) list; (* variables *)
  cons   : (string * string) list; (* algebraic data types *)
  tnames : (string * string) list; (* typenames *)
  tvars  : (string * string) list; (* type variables *)
}

type kind =
    Ident
  | Cons
  | Type
  | Tvar

type mapkind = kind * string * string

let proj kind strdata =
  match kind with
    | Ident -> strdata.idents
    | Cons -> strdata.cons
    | Type -> strdata.tnames
    | Tvar -> strdata.tvars
	
let initial = {
  idents = [];
  cons   = [];
  tnames = [];
  tvars  = [];
}

let merge x y = {
  idents = x.idents @ y.idents;
  cons   = x.cons @ y.cons;
  tnames = x.tnames @ y.tnames;
  tvars  = x.tvars @ y.tvars
}

let lookup kind env x =
  let map = proj kind env in
    if List.mem_assoc x map then
      Some (List.assoc x map)
    else
      None

let lookup_ident = lookup Ident

let lookup_cons = lookup Cons

let lookup_typename = lookup Type

let lookup_default kind env x =
  match lookup kind env x with
    | None -> x
    | Some y -> y
     
(* Extend the environment with a mapping from [key] to [fresh] *)
let add_map env mapping =
  match mapping with
    | (Ident, key, fresh) ->
	{ env with idents = (key, fresh) :: env.idents }
    | (Cons, key, fresh) ->
	{ env with cons = (key, fresh) :: env.cons }
    | (Type, key, fresh) ->
	{ env with tnames = (key, fresh) :: env.tnames }
    | (Tvar, key, fresh) ->
	{ env with tvars = (key, fresh) :: env.tvars }

(* Updates a core type using the mapping given by the environment *)
let update_core_type env typ =
  remap_core_type (lookup_default Type env) (lookup_default Tvar env) typ
 
let rec alpha_convert_structure_item env freshgen (item_desc, item_id) =
  let env, fresh, desc = match item_desc with
    | Iexternal(_, _) ->
	env, freshgen, item_desc
    | Itypedecl decls ->
	let env, fresh, decls = alpha_convert_typedecl env freshgen decls in
	  env, fresh, (Itypedecl decls)
  in
    env, fresh, (desc, item_id)
	 
and alpha_convert_pattern env freshgen (patt_desc, patt_id) =
  let env, freshgen, desc =
    match patt_desc with
      | Pany
      | Pconst _ ->
	  env, freshgen, patt_desc
      | Pvar v ->
	  let gen, fresh = Fresh.next freshgen in
	  let v'         = v^^(Fresh.var fresh) in
	  let env        = add_map env (Ident, v, v') in
	    env, gen, (Pvar v')
      | Ptuple patterns ->
	  let env, fresh, patterns = List.fold_right 
	    (fun patt (env, fresh, patterns) ->
	      let e,f,p = alpha_convert_pattern env fresh patt in
		(e, f, p :: patterns)
	    ) patterns (env, freshgen, [])
	  in
	    env, fresh, (Ptuple patterns)
      | Pconstruct(cons, None) ->
	  let cons' = match lookup_cons env cons with
	    | None -> cons
	    | Some x -> x
	  in
	    env, freshgen, (Pconstruct(cons', None))
      | Pconstruct(cons, Some patt) ->
	  let cons' = match lookup_cons env cons with
	    | None -> cons
	    | Some x -> x
	  in
	  let env, fresh, patt = alpha_convert_pattern env freshgen patt in
	    env, fresh, Pconstruct(cons', Some patt)
  in
    env, freshgen, (desc, patt_id)
      
and alpha_convert_matching env freshgen bindings =
  match bindings with
    | [] -> env, freshgen, []
    | (patt, expr) :: l ->
	let freshgen, binding =
	  let env', freshgen, patt = 
	    alpha_convert_pattern env freshgen patt in
	  let freshgen, expr = 
	    alpha_convert_expression env' freshgen expr 
	  in
	  let binding = (patt, expr) in
	    freshgen, binding
	in 
	let env, freshgen, bindings = 
	  alpha_convert_matching env freshgen l 
	in
	  env, freshgen, (binding :: bindings)

and alpha_convert_expression env freshgen (expr_desc, expr_id) =
  let freshgen, desc = match expr_desc with
    | Eident id ->
	let ident = match lookup_ident env id with
	  | None -> failwith ("no fresh instance for ident "^id)
	  | Some x -> x
	in
	  freshgen, (Eident ident)
    | Elet(var, expr, body) ->
	let freshgen, expr  = alpha_convert_expression env freshgen expr in
	let freshgen, fresh = Fresh.next freshgen in
	let var'            = var^^(Fresh.var fresh) in
	let env             = add_map env (Ident, var, var') in
	let freshgen, body  = alpha_convert_expression env freshgen body in
	  freshgen, (Elet(var', expr, body))
    | Efunction(var, body) ->
	let freshgen, fresh = Fresh.next freshgen in
	let var'            = var^^(Fresh.var fresh) in
	let env             = add_map env (Ident, var, var') in
	let freshgen, body  = alpha_convert_expression env freshgen body in
	  freshgen, (Efunction(var', body))
    | Efixpoint(vfix, var, body) ->
	let freshgen, fresh = Fresh.next freshgen in
	let vfix'           = vfix^^(Fresh.var fresh) in
	let env             = add_map env (Ident, vfix, vfix') in
	let freshgen, fresh = Fresh.next freshgen in
	let var'            = var^^(Fresh.var fresh) in
	let env             = add_map env (Ident, var, var') in
	let freshgen, body  = alpha_convert_expression env freshgen body in
	  freshgen, (Efixpoint(vfix', var', body))
    | Eapply(func, args) ->
	let freshgen, args = List.fold_right 
 	  (fun arg (freshgen, args) ->
	     let freshgen, arg = 
	       alpha_convert_expression env freshgen arg
	     in
	       (freshgen, arg :: args)
	  ) args (freshgen, []) in
	let freshgen, func =
	  alpha_convert_expression env freshgen func 
	in
	  freshgen, (Eapply(func, args))
    | Ematch(matched_expr, bindings) ->
	let freshgen, matched_expr =
	  alpha_convert_expression env freshgen matched_expr
	in
	let _, freshgen, bindings = 
	  alpha_convert_matching env freshgen bindings
	in
	  freshgen, (Ematch(matched_expr, bindings))
    | Etuple fields ->
	let freshgen, fields = List.fold_right 
	  (fun field (freshgen, fields) ->
	     let freshgen, field = 
	       alpha_convert_expression env freshgen field
	     in
	       (freshgen, field :: fields)
	  ) fields (freshgen, []) in
	  freshgen, (Etuple fields)
    | Econstruct(cons, None) ->
	let cons = match lookup_cons env cons with
	  | None -> cons
	  | Some x -> x
	in
	  freshgen, (Econstruct(cons, None))
    | Econstruct(cons, Some e) ->
	let cons = match lookup_cons env cons with
	  | None -> cons
	  | Some x -> x
	in
	let freshgen, e =
	  alpha_convert_expression env freshgen e
	in
	  freshgen, (Econstruct(cons, Some e))
    | Econstant _ ->
	freshgen, expr_desc
  in
    freshgen, (desc, expr_id)

and alpha_convert_typedecl env freshgen mutual_decls =

  let env, freshgen = List.fold_left (fun (env, freshgen) decl ->
    match decl.tdecl_kind with
      | SumType _ ->
	  let gen, fresh = Fresh.next freshgen in
	  let typename'  = decl.tdecl_name^^(Fresh.var fresh) in
	  let env        = add_map env (Type, decl.tdecl_name, typename') in
	    (env, gen)
      | Builtin ->
	  (env, freshgen)
  ) (env, freshgen) mutual_decls in
    
  let rec loop env freshgen mutual_decls =
    match mutual_decls with
      | [] -> env, freshgen, []
      | decl :: l ->
	  let env, freshgen, kind = 
	    match decl.tdecl_kind with
	      | SumType inductive_def ->
		  let env, freshgen, datacons = 
		    convert_inductive_def env freshgen inductive_def 
		  in
		    env, freshgen, (SumType datacons)
	      | Builtin ->
		  env, freshgen, Builtin
	  in
	  let env, freshgen, decls = loop env freshgen l in
	  let typename = lookup_default Type env decl.tdecl_name in
	  let decl = {
	    tdecl_name = typename;
	    tdecl_kind = kind
	  } in
	    env, freshgen, (decl :: decls)
  in
    loop env freshgen mutual_decls

and convert_inductive_def env freshgen inductive_def =
  match inductive_def with
    | IndAbs(var, body) ->
	let freshgen, fresh = Fresh.next freshgen in
	let var'            = var^^(Fresh.var fresh) in
	let env             = add_map env (Tvar, var, var') in
	let env, freshgen, body  = convert_inductive_def env freshgen body in
	  env, freshgen, (IndAbs(var', body))
    | IndIntros rules ->
	let env, freshgen, rules = convert_inductive_constructors env freshgen rules in
	  env, freshgen, (IndIntros rules)
		    

and convert_inductive_constructors env freshgen constructors =
  match constructors with
    | [] -> env, freshgen, []
    | (cons, None) :: l ->
	let gen,fresh = Fresh.next freshgen in
	let new_cons  = cons^^(Fresh.var fresh) in
	let env       = add_map env (Cons, cons, new_cons) in
	let env, freshgen, constructors = convert_inductive_constructors env gen l in
	  env, freshgen, ((new_cons, None) :: constructors)
    | (cons, Some domain) :: l ->
	let gen,fresh = Fresh.next freshgen in
	let new_cons  = cons^^(Fresh.var fresh) in
	let env       = add_map env (Cons, cons, new_cons) in
	let domain    = update_core_type env domain in
	let env, freshgen, constructors = convert_inductive_constructors env gen l in
	  env, freshgen, ((new_cons, Some domain) :: constructors)

let alpha_convert_program env freshgen (decls, main) =
  let env, freshgen, decls = List.fold_left (fun (env, freshgen, decls) decl ->
    let env, freshgen, decl = alpha_convert_structure_item env freshgen decl in
      (env, freshgen, decl :: decls)
  ) (env, freshgen, []) decls
  in
  let freshgen, main = alpha_convert_expression env freshgen main in
    (env, freshgen, (List.rev decls, main))

let alpha_conversion freshgen program =
  let _, freshgen, program = alpha_convert_program initial freshgen program in
    freshgen, program
