(* Ast module. *)

(* The [Ast] module defines the abstract syntax tree of our flavor
of core ML. It is basically a pure call-by-value simply-typed lambda-calculus extended with~:

\begin{itemize}
\item algebraic data types
\item tuples
\item pattern matching as a general elimination form for the aforementioned
  data
\item \textit{let}-definitions
\item a fixpoint operator
\item some builtin types
\end{itemize}

The compiler itself is written in a pure, call-by-value subset of OCaml.

A program is structured as a list of type definitions, in the style
of Caml, followed by a possibly empty list of value definitions, and ending
with a \textbf{main} expression. During parsing (cf. [Parsing]), all these definitions are
collapsed into one big expression. The types and terms are in first-order
representation, care must be taken to alpha-convert them before playing with
substitutions.

As can be seen from the [pattern] and [expr] definition, each node
carries an id (which should be unique), allowing us to annotate
the AST using simple maps.

*)

(*i*)
open Utils
open Printf
(*i*)

type unique_id = int
type vvar = string
type ident = vvar
type datacons = string
type typecons = string

(*i Constants i*)
type constant =
  | CInt of int
  | CChar of char
  | CFloat of float

(*i Patterns i*)
type pattern = (patt_desc * unique_id)
and patt_desc =
  | Pany
  | Pvar of vvar
  | Pconst of constant
  | Ptuple of pattern list
  | Pconstruct of datacons * pattern option

(*i Expressions i*)
type expr = (expr_desc * unique_id)
and expr_desc =
  | Eident of ident
  | Econstant of constant
  | Elet of vvar * expr * expr
  | Efunction of vvar * expr
  | Efixpoint of vvar * vvar * expr
  | Eapply of expr * expr list
  | Ematch of expr * matching
  | Etuple of expr list
  | Econstruct of datacons * expr option
and matching = (pattern * expr) list

(*i Parsed types i*)
type core_type = (typecons, string) Terms.t

type inductive_type =
  | IndAbs of string * inductive_type
  | IndIntros of (datacons * core_type option) list

type type_declaration =
    { tdecl_name : string;
      tdecl_kind : type_kind }
and type_kind =
  | Builtin
  | SumType of inductive_type

(*i Items. i*)
(* Note that type declarations can be mutually recursive. *)
type item = (item_desc * unique_id)
and item_desc =
  | Iexternal of vvar * core_type
  | Itypedecl of type_declaration list

type decls = item list

type program = decls * expr


(*i****************************************************************i*)

(*s During the process of compilation, we will often have to manipulate
the binders (for instance during alpha-conversion). In order to be
able to have an intelligible result, we manage to keep the original
identifier in prefix position. The "salt" is separated of the prefix
by [sep_char], a constant and special-purpose character.

This allows us to still print unmangled identifiers after all the compilation
passes. *)

let sep_char  = char_of_int 255
let separator = String.make 1 sep_char

let (^^) a b = a^separator^b

let printable_clean x =
  let x_len = String.length x in
  let rec prefix i acc =
    if i = x_len || x.[i] == sep_char then
      acc
    else
      prefix (i+1) (acc^(String.make 1 x.[i]))
  in
    prefix 0 ""

let printable_dirty x = 
  let x' = String.copy x in
    for i = 0 to (String.length x') - 1 do
      if x'.[i] = sep_char then
	x'.[i] <- '_'
      else
	()
    done;
    x'

let printable = printable_clean


(*i****************************************************************i*)

(*s These are often used functions on the types, e.g. application of
  a polymorphic type to a type argument, or extraction of the introduction
  rules from a declaration. *)

(*i inductive[var\arg] i*)
let rec inductive_subst var arg inductive =
  match inductive with
    | IndAbs(var', body) ->
	if var = var' then
	  inductive
	else
	  IndAbs(var', inductive_subst var arg body)
    | IndIntros rules ->
	let rules = List.map (function
	  | (cons, None) -> (cons, None)
	  | (cons, Some domain) ->
	      (cons, Some (Terms.term_subst var arg domain))
	) rules
	in
	  IndIntros rules

let inductive_apply indtype typearg =
  match indtype with
    | IndAbs(s, body) ->
	inductive_subst s typearg body
    | IndIntros _ ->
	failwith "Ast.inductive_apply : type is not an abstraction."

let rec constructors_of_inductive_def indtype =
  match indtype with
    | IndAbs(_, body) ->
	constructors_of_inductive_def body
    | IndIntros rules -> rules
	

(*i****************************************************************i*)

(*s These are often used functions on the expressions. *)

let desc e = fst e

let variables_of_pattern patt =
  let rec aux acc (patt_desc, _) =
    match patt_desc with
      | Pany
      | Pconst _
      | Pconstruct(_, None) -> acc
      | Pvar vvar -> vvar :: acc
      | Ptuple patts ->
	  List.fold_left aux acc patts
      | Pconstruct(_, Some patt) ->
	  aux acc patt
  in
  let vars = aux [] patt in
  let cond = Utils.check_uniqueness (=) (List.sort String.compare vars) in
    if cond then
      vars
    else
      failwith "Ast.variables_of_pattern : variables can't be bound multiple times in patterns."

let variables_of_pattern_with_ids patt =
  let rec aux acc (patt_desc, patt_id) =
    match patt_desc with
      | Pany
      | Pconst _
      | Pconstruct(_, None) -> acc
      | Pvar vvar -> (vvar, patt_id) :: acc
      | Ptuple patts ->
	  List.fold_left aux acc patts
      | Pconstruct(_, Some patt) ->
	  aux acc patt
  in
  let vars = aux [] patt in
  let vars' = List.map fst vars in
  let cond = Utils.check_uniqueness (=) (List.sort String.compare vars') in
    if cond then
      vars
    else
      failwith "Ast.variables_of_pattern : variables can't be bound multiple times in patterns."


let remap_pattern patt f =
  let rec aux (patt_desc, patt_id) =
    let desc = match patt_desc with
      | Pany
      | Pconst _
      | Pconstruct(_, None) -> 
	  patt_desc
      | Pvar v ->
	  Pvar (f v)
      | Ptuple patts ->
	  Ptuple (List.map aux patts)
      | Pconstruct(cons, Some patt') ->
	  Pconstruct(cons, Some (aux patt'))
    in
      (desc, patt_id)
  in aux patt

let remap_core_type = Terms.remap_term

let rec is_value (expr_desc, _) =
  match expr_desc with
    | Eident _
    | Econstant _
    | Efunction(_, _)
    | Efixpoint(_, _, _)
    | Etuple _
    | Econstruct(_, _) -> true
    | Elet(_, _, body) ->
	is_value body
    | _ -> false

(*i e[x\y] i*)
let rec subst x y (expr_desc, expr_id) =
  let desc = match expr_desc with
    | Eident x' ->
	if x = x' then y else expr_desc
    | Econstant _ -> 
	expr_desc
    | Elet(v, bound, body) ->
	let bound' = subst x y bound in
	  if v = x then
	    Elet(v, bound, body)
	  else
	    Elet(v, bound', subst x y body)
    | Efunction(v, body) ->
	if v = x then
	  expr_desc
	else
	  Efunction(v, subst x y body)
    | Efixpoint(fixv, v, body) ->
	if x = fixv || x = v then
	  expr_desc
	else
	  Efixpoint(fixv, v, subst x y body)
    | Eapply(func, args) ->
	Eapply(subst x y func, List.map (subst x y) args)
    | Ematch(e, matching) ->
	let e' = subst x y e in
	let matching' = List.map (fun (patt, expr) ->
	  if List.mem x (variables_of_pattern patt) then
	    (patt, expr)
	  else
	    (patt, subst x y expr)
	) matching in
	  Ematch(e', matching')
    | Etuple fields ->
	Etuple (List.map (subst x y) fields)
    | Econstruct(_, None) ->
	expr_desc
    | Econstruct(cons, Some expr) ->
	Econstruct(cons, Some (subst x y expr))
  in
    (desc, expr_id)

let free_variables expr =
  let rec fvs env acc (expr_desc, expr_id) =
    match expr_desc with
      | Eident x ->
	  if List.mem x env then
	    acc
	  else
	    ((x, expr_id) :: acc)
      | Econstant _ -> 
	  acc
      | Elet(v, bound, body) ->
	  let acc = fvs env acc bound in
	    fvs (v :: env) acc body
      | Efunction(v, body) ->
	  fvs (v :: env) acc body
      | Efixpoint(fixv, v, body) ->
	  fvs (fixv :: v :: env) acc body
      | Eapply(func, args) ->
	  let acc = fvs env acc func in
	    List.fold_left (fvs env) acc args
      | Ematch(e, matching) ->
	  let acc = fvs env acc e in
	    List.fold_left (fun acc (patt, expr) ->
	      let env = List.rev_append (variables_of_pattern patt) env in
		fvs env acc expr
	    ) acc matching
      | Etuple fields ->
	  List.fold_left (fvs env) acc fields
      | Econstruct(_, None) ->
	  acc
      | Econstruct(cons, Some expr) ->
	  fvs env acc expr
  in
    fvs [] [] expr

(*i**************************************************************i*)

(*s The following functions are used to assign unique ids to each node
of an expression. Each function is parameterized by a fresh variable
generator [freshgen]. The functions are hand-written in monadic style,
threading the fresh variable generator during the recursion. The code
may seem quite messy to the programmer for which the respective usages
of [fold_left] and [fold_right] is not clear \ldots 

Please note that any map relying on ids will become useless, since the
key for a given node will not be the same after refresh.
*)

let rec refresh_pattern freshgen (patt_desc, patt_id) =
  let desc, freshgen = match patt_desc with
    | Ptuple patts ->
	let patts, freshgen = List.fold_right (fun p (patts, gen) ->
	  let p, gen = refresh_pattern gen p in
	    (p :: patts, gen)
	) patts ([], freshgen) in
	  (Ptuple patts), freshgen
    | Pconstruct(cons, Some p) ->
	let p, freshgen = refresh_pattern freshgen p in
	  (Pconstruct(cons, Some p)), freshgen
    | _ -> patt_desc, freshgen
  in
  let freshgen, id = Fresh.next freshgen in
    (desc, id), freshgen


let rec refresh_expression freshgen (expr_desc, expr_id) =
  let desc, freshgen = match expr_desc with
    | Elet(var, bound, body) ->
	let bound, freshgen = refresh_expression freshgen bound in
	let body, freshgen  = refresh_expression freshgen body in
 	  (Elet(var, bound, body)), freshgen
    | Efunction(vvar, body) ->
	let body, freshgen = refresh_expression freshgen body in
	  (Efunction(vvar, body)), freshgen
    | Efixpoint(vfix, vvar, body) ->
	let body, freshgen = refresh_expression freshgen body in
	  (Efixpoint(vfix, vvar, body)), freshgen
    | Eapply(func, args) ->
	let func, freshgen = refresh_expression freshgen func in
	let args, freshgen = List.fold_right (fun arg (args, freshgen) ->
	  let arg, freshgen = refresh_expression freshgen arg in
	    (arg :: args, freshgen)
	) args ([], freshgen) 
	in
	  (Eapply(func, args)), freshgen
    | Ematch(matched_expr, matching) ->
	let matched_expr, freshgen = refresh_expression freshgen matched_expr in
	let matching, freshgen = List.fold_right (fun (patt, branch) (matching, freshgen) ->
	  let patt, freshgen   = refresh_pattern freshgen patt in
	  let branch, freshgen = refresh_expression freshgen branch in
	    ((patt, branch) :: matching, freshgen)
	) matching ([], freshgen) 
	in
	  (Ematch(matched_expr, matching)), freshgen
    | Etuple fields ->
	let fields, freshgen = List.fold_right (fun field (fields, freshgen) ->
	  let field, freshgen = refresh_expression freshgen field in
	    (field :: fields, freshgen)
	) fields ([], freshgen)
	in
	  (Etuple fields), freshgen
    | Econstruct(datacons, Some e) ->
	let e, freshgen = refresh_expression freshgen e in
	  (Econstruct(datacons, Some e)), freshgen
    | Eident _
    | Econstant _
    | Econstruct(_, None) ->
	expr_desc, freshgen
  in
  let freshgen, id = Fresh.next freshgen in
    (desc, id), freshgen


let rec refresh_structure_item freshgen (item_desc, item_id) =
  let freshgen, id = Fresh.next freshgen in
    (item_desc, id), freshgen


let rec refresh_program freshgen (decls, main) =
  let freshgen, decls = List.fold_right (fun elt (freshgen, items) ->
    let item, freshgen = refresh_structure_item freshgen elt in
      (freshgen, item :: items)
  ) decls (freshgen, []) in
  let expr, freshgen = refresh_expression freshgen main in
    ((decls, expr), freshgen)
      


(*i****************************************************************i*)

(*s Some helpful functions to create expressions. Note that when
using theses, the id of each node is set to $-1$. Thus, a call
to [refresh_program] may be helpful after a compilation pass making
use of theses. *)

let mkexpr e = (e, -1)

let mkident v = mkexpr (Eident v)

let mkfun v body =
  mkexpr (Efunction(v, body))

let mkfix fix v body =
  mkexpr (Efixpoint(fix,v, body))

let mkapp func arg =
  mkexpr (Eapply(func, [arg]))

let mklet bindpatt bound body =
  mkexpr (Elet(bindpatt, bound, body))

let mkconst c =
  mkexpr (Econstant c)

let mkmatch e matching =
  mkexpr (Ematch(e, matching))

let mktuple fields =
  mkexpr (Etuple fields)

let mkconstruct cons  expr =
    mkexpr (Econstruct(cons, expr))


let mkpatt p = (p, -1)

let mkpany = mkpatt Pany

let mkpvar v = mkpatt (Pvar v)

let mkpconst c = mkpatt (Pconst c)

let mkpconstruct cons patt = mkpatt (Pconstruct(cons, patt))

let mkptuple fields = mkpatt (Ptuple fields)

(*i****************************************************************i*)

(*s The usual printing function. Note that identifiers are demangled
using the [printable] function. *)

let print_constant =
  function
    | CInt i -> string_of_int i
    | CChar c -> String.make 1 c
    | CFloat f -> string_of_float f

let rec print_pattern (patt_desc, patt_id) =
  let s = match patt_desc with
    | Pany -> "_"
    | Pvar v -> (printable v)
    | Pconst c -> print_constant c
    | Ptuple patts ->
	"("^(print_list "," print_pattern patts)^")"
    | Pconstruct(datacons, None) ->
	(printable datacons)
    | Pconstruct(datacons, Some patt) ->
	(printable datacons)^" "^(print_pattern patt)
  in
    s(*^"$"^(string_of_int patt_id)*)

let rec print_expr (expr_desc, expr_id) =
  let s = match expr_desc with
    | Eident i -> (printable i)
    | Econstant c ->
	print_constant c
    | Elet(v, bound, body) ->
	sprintf "let %s = %s in\n%s" 
	  (printable v) (print_expr bound) (print_expr body)
    | Efunction(v, body) ->
	sprintf "(fun %s => %s)" (printable v) (print_expr body)
    | Efixpoint(fixv, v, body) ->
	sprintf "(fix %s %s => %s)" (printable fixv) (printable v) (print_expr body)
    | Eapply(func, args) ->
	let f    = print_expr func in
	let args = print_list " " print_expr args in
	  sprintf "(%s %s)" f args
    | Ematch(e, matching) ->
	let e = print_expr e in
	let m = print_matching matching in
	  sprintf "match %s with [ %s ]" e m
    | Etuple fields ->
	"("^(print_list "," print_expr fields)^")"
    | Econstruct(datacons, None) ->
	printable datacons
    | Econstruct(datacons, Some e) ->
	"("^(printable datacons)^" "^(print_expr e)^")"
  in
    s(*^"$"^(string_of_int expr_id)*)

and print_matching m =
  print_list "\n| " (fun (patt, e) ->
    sprintf "%s => %s" (print_pattern patt) (print_expr e)
  ) m

let rec print_core_type typ =
  match typ with
    | Terms.TermVar v -> v
    | Terms.TermCons("*", fields) ->
	"("^(print_list " * " print_core_type fields)^")"
    | Terms.TermCons("->", [dom; codom]) ->
	sprintf "(%s -> %s)" (print_core_type dom) (print_core_type codom)
    | Terms.TermCons(cons, []) ->
	printable cons
    | Terms.TermCons(cons, fields) ->
	let fields = print_list ", " print_core_type fields in
	  (printable cons)^"("^fields^")"

let rec print_inductive inductive =
  match inductive with
    | IndAbs(arg, body) ->
	sprintf "abs %s. %s" arg (print_inductive body)
    | IndIntros rules ->
	let cons = print_list "\n| " (fun (datacons, opt_typ) ->
	  match opt_typ with
	    | None -> (printable datacons)
	    | Some typ ->
		(printable datacons)^" :"^(print_core_type typ)
	) rules in
	  "["^cons^"]"

let print_typedecl decl =
  let name = printable decl.tdecl_name in
  let kind = match decl.tdecl_kind with
    | Builtin -> "builtin"
    | SumType inductive ->
	print_inductive inductive
  in
    sprintf "type %s = %s;\n" name kind

let print_item (item_desc, _) =
  match item_desc with
    | Iexternal(name, typ) ->
	sprintf "external %s : %s;\n" name (print_core_type typ)
    | Itypedecl decls ->
        let s = print_list "\nand " print_typedecl decls in
	  s^";\n"

let print_decls decls = List.iter (fun x -> printf "%s\n" (print_item x)) decls

let print_program (decls, main) =
  print_decls decls;
  printf "%s\n" (print_expr main)

