(* Closure conversion *)

(*
[fun x => e : Tx -> Te] = 
    let F = fun (c,x) => 
      let fv1 = field1(c) in
      let fv2 = field2(c) in
        ...
      let fvn = fieldn(c) in
        [e]
    in
      Closure(F, fv1, ..., fvn)

: muX.((X * Tx) -> [[B]], d1 * ... * dn)

type t ('tx,'tb,'tfv) =
 | Closure of ((t * 'tx) -> 'tb) * 'tfv

[a b] = avec (a : t -> u) et (b : t)
    clos : mu X.(((X * [t]) -> [u]) * d1 * ... * dn)
    let clos = [a] in    
    field0(clos) (clos, [b]) *)

(*i*)
open Ast
open Printf
open Types.NoDebug
open Types
open Terms
(*i*)

(*s The input signature for the closure conversion functor contains
  a typing function [id_typing] mapping the nodes' unique ids to their \textbf{monomorphic}
  type.
*)

module type InputSig =
sig

  val id_typing : Ast.unique_id -> Types.NoDebug.term

end

module Make(I : InputSig) =
struct

  let stub_cons = "__MAO_Stub"
  
  let apply fn arg =
    let mao_fn   = mkident "__mao_fn" in
    let mao_code = mkident "__mao_code" in
    let action   = mkapp mao_code (mktuple [mao_fn; arg]) in
    let pattern = 
      mkpconstruct "__MAO_Closure" (Some (mkptuple [mkpvar "__mao_code"; mkpany]))
    in
       mklet "__mao_fn" fn (mkmatch mao_fn [ (pattern, action) ])

  let closconv_function fun_arg body id free_vars =
    match free_vars with
      | [] ->
	  let new_arg     = "__mao_argument" in	
	  let arg_pattern = mkptuple [mkpany; mkpvar fun_arg] in
	  let func = mkfun new_arg
	    (mkmatch (mkident new_arg)
	      [ (arg_pattern, body) ]) in
	  let closure = 
	    mkconstruct "__MAO_Closure" (Some (mktuple [func; mkconstruct stub_cons None]))
	  in
	    (closure, [])
      | _ ->
	  let free_vars = List.sort (fun (x, _) (y, _) -> String.compare x y) free_vars in
	  let free_vars = Utils.filter_duplicates (fun (x, _) (y, _) -> x = y) free_vars in
	  let free_vars = List.map (fun (var, var_id) -> (var, I.id_typing var_id)) free_vars in
	  let fv_expr, fv_patt =
	    match free_vars with
	      | [] -> failwith "impossible case"
	      | [(fv, _)] ->
		  let fv_expr = mkident fv in
		  let fv_patt = mkpvar fv in
		    fv_expr, fv_patt
	      | _ ->
		  let fv_expr = mktuple (List.map (fun (var, _) -> mkident var) free_vars) in
		  let fv_patt = mkptuple (List.map (fun (var, _) -> mkpvar var) free_vars) in
		    fv_expr, fv_patt
	  in
	  let constr    = sprintf "__MAO_env_%d" id in
	  let new_arg   = "__mao_arg" in
	  let env_var   = "__mao_env" in
	  let clos_var  = "__mao_clos" in
	  let fv_intro  = mkconstruct constr (Some fv_expr) in
	  let fv_elim   = 
	    mkmatch (mkident env_var) 
	      [(mkpconstruct constr (Some fv_patt), body)]
	  in
	  let clos_elim =
	    mkmatch (mkident clos_var)
	      [(mkpconstruct "__MAO_Closure" (Some (mkptuple [mkpany; mkpvar env_var])), fv_elim)]
	  in
	  let arg_elim =
	    mkmatch (mkident new_arg)
	      [(mkptuple [mkpvar clos_var; mkpvar fun_arg], clos_elim)]
	  in
	  let func = mkfun new_arg arg_elim in
	  let closure = mkconstruct "__MAO_Closure" (Some (mktuple [func; fv_intro])) in
	    (closure, [(constr, List.map snd free_vars)])


  let closconv_fixpoint fixvar fun_arg body id free_vars =
    match free_vars with
      | [] ->
	  let new_arg     = "__mao_argument" in	
	  let arg_pattern = mkptuple [mkpany; mkpvar fun_arg] in
	  let fixpoint_closure = 
	    mkconstruct "__MAO_Closure" (Some (mktuple [mkident fixvar; mkconstruct stub_cons None]))
	  in
	  let body = Ast.subst fixvar (desc fixpoint_closure) body in
	  let func = mkfix fixvar new_arg (mkmatch (mkident new_arg) [ (arg_pattern, body) ]) in
	  let closure =
	    mkconstruct "__MAO_Closure" (Some (mktuple [func; mkconstruct stub_cons None]))
	  in
	    (closure, [])
      | _ ->
	  let free_vars = List.sort (fun (x, _) (y, _) -> String.compare x y) free_vars in
	  let free_vars = Utils.filter_duplicates (fun (x, _) (y, _) -> x = y) free_vars in
	  let free_vars = List.map (fun (var, var_id) -> (var, I.id_typing var_id)) free_vars in
	  let fv_expr, fv_patt =
	    match free_vars with
	      | [] -> failwith "impossible case"
	      | [(fv, _)] ->
		  let fv_expr = mkident fv in
		  let fv_patt = mkpvar fv in
		    fv_expr, fv_patt
	      | _ ->
		  let fv_expr = mktuple (List.map (fun (var, _) -> mkident var) free_vars) in
		  let fv_patt = mkptuple (List.map (fun (var, _) -> mkpvar var) free_vars) in
		    fv_expr, fv_patt
	  in
	  let constr    = sprintf "__MAO_env_%d" id in
	  let new_arg   = "__mao_arg" in
	  let env_var   = "__mao_env" in
	  let clos_var  = "__mao_clos" in
	  let fv_intro  = mkconstruct constr (Some fv_expr) in
	  let fixpoint_closure = 
	    mkconstruct "__MAO_Closure" (Some (mktuple [mkident fixvar; mkident env_var]))
	  in
	  let body      = Ast.subst fixvar (desc fixpoint_closure) body in
	  let fv_elim   = 
	    mkmatch (mkident env_var) 
	      [(mkpconstruct constr (Some fv_patt), body)]
	  in
	  let clos_elim =
	    mkmatch (mkident clos_var)
	      [(mkpconstruct "__MAO_Closure" (Some (mkptuple [mkpany; mkpvar env_var])), fv_elim)]
	  in
	  let arg_elim =
	    mkmatch (mkident new_arg)
	      [(mkptuple [mkpvar clos_var; mkpvar fun_arg], clos_elim)]
	  in
	  let func = mkfix fixvar new_arg arg_elim in
	  let closure = mkconstruct "__MAO_Closure" (Some (mktuple [func; fv_intro])) in
	    (closure, [(constr, List.map snd free_vars)])


  let rec closconv env (expr_desc, expr_id) =
    let desc, env = match expr_desc with
      | Eident _
      | Econstant _
      | Econstruct(_, None) ->
	  expr_desc, env
      | Elet(var, bound, body) ->
	  let bound, env = closconv env bound in
	  let body,  env = closconv env body in
	    Elet(var, bound, body), env
      | Efunction(var, body) ->
	  let free_vars  = Ast.free_variables (expr_desc, expr_id) in
	  let body, env  = closconv env body in
	  let code, decl = closconv_function var body expr_id free_vars in
	    (desc code), (decl @ env)
      | Efixpoint(vfix, var, body) ->
	  let free_vars  = Ast.free_variables (expr_desc, expr_id) in
	  let body, env  = closconv env body in
	  let code, decl = closconv_fixpoint vfix var body expr_id free_vars in
	  let env        = decl @ env in
	    (desc code), (decl @ env)
      | Eapply(func, args) ->
	  let func, env   = closconv env func in
	  let result, env = List.fold_left (fun (result, env) arg ->
	    let arg, env = closconv env arg in
	      (apply result arg, env)
	  ) (func, env) args in
	    desc result, env
      | Ematch(matched_expr, matching) ->
	  let matched_expr, env = closconv env matched_expr in
	  let matching, env = List.fold_right
	    (fun (pattern, action) (matching, env) ->
	      let action, env = closconv env action in
		((pattern, action) :: matching, env)
	    ) matching ([], env)
	  in
	    Ematch(matched_expr, matching), env
      | Etuple fields ->
	  let fields, env = List.fold_right (fun elt (fields, env) ->
	    let elt, env = closconv env elt in
	      (elt :: fields, env)
	  ) fields ([], env) in
	    (Etuple fields), env
      | Econstruct(datacons, Some e) ->
	  let e, env = closconv env e in
	    Econstruct(datacons, Some e), env
    in
      (desc, expr_id), env

  let closure_type domain codomain =
    TermCons("__MAO_closure", [domain; codomain])
      
  let rec closify_type typ =
    match typ with
      | TermVar _ -> typ
      | TermCons("->", [domain; codomain]) ->
	  closure_type (closify_type domain) (closify_type codomain)
      | TermCons("->", _) ->
	  failwith "inconsistent arrow type in closure conversion"
      | TermCons(tc, subtypes) ->
	  TermCons(tc, List.map closify_type subtypes)

  let closify_kind kind =
    match kind with
      | Builtin -> kind
      | SumType (IndIntros introrules) ->
	  let intros = List.map (function
	    | (datacons, None) -> 
		(datacons, None)
	    | (datacons, Some typ) ->
		(datacons, Some (closify_type typ))
	  ) introrules in
	    SumType (IndIntros intros)
      | SumType IndAbs(_, _) ->
	  failwith "Closure : bug found in closify_kind"

  let closify_decl decl =
    { decl with
      tdecl_kind = closify_kind decl.tdecl_kind }

  let rec closify_item (it_desc, it_id) =
    let desc = match it_desc with
      | Iexternal(_, _) ->
	  it_desc
      | Itypedecl decls ->
	  Itypedecl (List.map closify_decl decls)
    in
      (desc, it_id)


  let pseudo_cast =
    Terms.remap_term
      (fun x -> x)
      (fun _ -> failwith "Closure.pseudo_cast : bug found, type variable in inductive definition.")

  let closure_conversion (decls, main) =
    let main, new_decls = closconv [] main in
    let new_decls = List.sort (fun (x,_) (y,_) -> String.compare x y) new_decls in
    let new_decls = Utils.filter_duplicates (fun (x,_) (y,_) -> x = y) new_decls in
    let sumtype = List.map (fun (cons, domain) ->
      let domtype = match domain with
	| [] -> None
	| [dom] -> 
	    Some (pseudo_cast dom)
	| _ ->
	    let domain = tuple_type domain in
	    let domain = pseudo_cast domain in
	      Some domain
      in
	(cons, domtype)
    ) new_decls in
    let sumtype = (stub_cons, None) :: sumtype in
    let env_decl = { 
      tdecl_name = "__MAO_environments";
      tdecl_kind = SumType (IndIntros sumtype)
    } in
    let env_decl = closify_decl env_decl in
      (*
	type closure ('a, 'b) = 
	[ Closure of (((closure ('a, 'b))  * 'a) -> 'b) * foo ]
      *)
    let clos_domtype =
      tuple_type
	[arrow_type
	  (tuple_type [(mktype "__MAO_closure" [mkvar "a0"; mkvar "a1"]); mkvar "a0"])
	  (mkvar "a1");
	 mk0ary "__MAO_environments"]
    in
    let clos_decl = {
      tdecl_name = "__MAO_closure";
      tdecl_kind = 
	SumType (IndAbs("a0",IndAbs("a1",IndIntros [("__MAO_Closure", Some clos_domtype)])))
    } in
    let decls = List.map closify_item decls in
    let decls = (Itypedecl [env_decl], 0) :: (Itypedecl [clos_decl], 0) :: decls in
      (decls, main)
	
end
