
(* ASTs in locally nameless notation. This representation
 * allows to avoid variable capture while doing
 * substitutions. *)

open Ast
open Utils
open Printf

type level = int
type arity = int

type dexpr =
  | DBident of level
  | DBfree of string
  | DBconstant of constant
  | DBlet of dexpr * dexpr
  | DBfunction of dexpr
  | DBfixpoint of dexpr
  | DBapply of dexpr * dexpr list
  | DBmatch of dexpr * dmatching
  | DBtuple of dexpr list
  | DBconstruct of datacons * dexpr option

(* We re-use the same patterns, just ignoring the variables
 * when evaluating. *)
and dmatching = (pattern * dexpr) list


let find_index var list =
  let rec aux i l =
    match l with
      | [] ->
	  None
      | v :: l ->
	  if var = v then
	    Some i
	  else
	    aux (i + 1) l
  in
    aux 0 list

let rec from_expr vars (expr, _) =
  match expr with
    | Eident v ->
	(match find_index v vars with
	  | None   -> DBfree v
	  | Some i -> DBident i)
    | Econstant c ->
	DBconstant c
    | Elet(v, bound, body) ->
	DBlet(from_expr vars bound,
	      from_expr (v :: vars) body)
    | Efunction(v, body) ->
	DBfunction(from_expr (v :: vars) body)
    | Efixpoint(fixv, v, body) ->
	DBfixpoint(from_expr (v :: fixv :: vars) body)
    | Eapply(func, args) ->
	DBapply(from_expr vars func,
	        List.map (from_expr vars) args)
    | Ematch(matched, matching) ->
	let matching = from_matching vars matching in
	let matched  = from_expr vars matched in
	  DBmatch(matched, matching)
    | Etuple fields ->
	let fields = List.map (from_expr vars) fields in
	  DBtuple fields
    | Econstruct(cons, None) ->
	DBconstruct(cons, None)
    | Econstruct(cons, Some e) ->
	DBconstruct(cons, Some (from_expr vars e))


and from_matching vars matching =
  List.map (fun (pattern, action) ->
    (* Warning : the indexing is done according to
     * the tree run of variables_of_pattern; that is
     * depth-first, left-to-right. Cf [Ast] module. *)
    let pattvars = variables_of_pattern pattern in
      (pattern, from_expr (pattvars @ vars) action)
  ) matching


let rec is_instance_of env matched (pattern, _) =
  match matched, pattern with
    | _, Pany -> Some env
    | _, (Pvar _) -> Some (matched :: env)
    | DBconstant c, Pconst c' ->
	if c = c' then
	  Some env
	else
	  None
    | DBtuple fields, Ptuple patts ->
	List.fold_left (fun env (field, patt) ->
	  match env with
	    | None -> None
	    | Some env ->
		is_instance_of env field patt
	) (Some env) (List.combine fields patts)
    | DBconstruct(cons, None), Pconstruct(cons', None) ->
	if cons = cons' then
	  Some env
	else
	  None
    | DBconstruct(cons, Some e), Pconstruct(cons', Some e') ->
	if cons = cons' then
	  is_instance_of env e e'
	else
	  None
    | _ -> None



let rec print e =
  match e with
    | DBident i ->
	Printf.sprintf "[%d]" i
    | DBfree s ->
	Printf.sprintf "<%s>" (Ast.printable s)
    | DBconstant c ->
	(match c with
	  | CInt i ->
	      string_of_int i
	  | CChar c ->
	      String.make 1 c
	  | CFloat f ->
	      string_of_float f)
    | DBlet(bound, body) ->
	Printf.sprintf "let = %s in\n%s" (print bound) (print body)
    | DBfunction e ->
	Printf.sprintf "(fun -> %s)" (print e)
    | DBfixpoint e ->
	Printf.sprintf "(fix -> %s)" (print e)
    | DBconstruct(cons, None) ->
	Ast.printable cons
    | DBconstruct(cons, Some e) ->
	Printf.sprintf "(%s(%s))" (Ast.printable cons) (print e)
    | DBapply(f, args) ->
	let f   = print f in
	let app = List.fold_left (fun acc arg -> "("^acc^(print arg)^")") f args in
	  app
    | DBtuple fields ->
	let args = Utils.print_list "," print fields in
	  "("^args^")"
    | DBmatch(e, matching) ->
	let e = print e in
	let m = print_matching matching in
	  sprintf "match %s with [ %s ]" e m

and print_matching m =
  print_list "\n|" (fun (patt, e) ->
    sprintf "%s => %s" (Ast.print_pattern patt) (print e)
  ) m

let rec subst x i e = (* [x/[i]]e, assuming x is locally closed *)
  match e with
    | DBident i' ->
	if i = i' then
	  x
	else
	  e
    | DBfree _
    | DBconstant _
    | DBconstruct(_, None) -> e
    | DBfunction body ->
	DBfunction (subst x (i+1) body)
    | DBfixpoint body ->
	DBfixpoint (subst x (i+2) body)
    | DBconstruct(label, Some e) ->
	DBconstruct(label, Some (subst x i e))
    | DBlet(bound, body) ->
	DBlet(subst x i bound, subst x (i+1) body)
    | DBapply(func, args) ->
	DBapply(subst x i func,
	        List.map (subst x i) args)
    | DBmatch(matched, matching) ->
	DBmatch(subst x i matched,
	        subst_matching x i matching)
    | DBtuple fields ->
	DBtuple (List.map (subst x i) fields)

and subst_matching x i matching =
  List.map (fun (pattern, action) ->
    let vars = Ast.variables_of_pattern pattern in
    let len  = List.length vars in
      (pattern, subst x (i+len) action)
  ) matching

      
(* big-step call-by-value reduction for De-Bruijn terms. *)
let rec eval expr =
  match expr with
    | DBident _ ->
	failwith "arghgh, dangling De Bruijn variable found"
    | DBfree _
    | DBconstant _
    | DBfunction _
    | DBfixpoint _
    | DBconstruct(_, None) ->
	expr
    | DBlet(bound, body) ->
	let bound = eval bound in
	let body  = subst bound 0 body in
	  eval body
    | DBapply(func, args) ->
	let args = List.map eval args in
	  eval_app func args
    | DBmatch(matched, matching) ->
	eval_match matched matching
    | DBtuple fields ->
	DBtuple (List.map eval fields)
    | DBconstruct(cons, Some e) ->
	DBconstruct(cons, Some (eval e))
	  
and eval_app func args =
  match args with
    | [] ->
	eval func
    | x :: l ->
	let nf = eval func in
	  (match nf with
	    | DBfunction body ->
		let body = subst x 0 body in
		  eval_app body l
	    | DBfixpoint body ->
		let body = subst nf 1 body in
		let body = subst x 0 body in
		  eval_app body l
	    | _ -> failwith (sprintf "runtime error in eval_app with %s" (print (DBapply(func, args)))))

and eval_match matched matching =
  let matched = eval matched in
  let rec do_match cases =
    match cases with
      | [] ->
	  failwith "failed matching"
      | (pattern, action) :: l ->
	  (match is_instance_of [] matched pattern with
	    | None ->
		do_match l
	    | Some env ->
		let (result, _) = List.fold_left (fun (expr, lvl) elt ->
		  (subst elt lvl expr, lvl+1)
		) (action, 0) env in
		  eval result
	  )
  in
    do_match matching


(* let f =  *)
(*   mkfun "x" (mkconstruct "Test" (Some (mkconstruct "Foo" (Some (mkident "x"))))) *)

(* let f = mkfun "x" (mkapp (mkident "x") (mkident "x")) *)

(* let omega = mkapp f f *)

(* let test = mkapp f (mkconst (CInt 3)) *)

(*  let _ = *)
(*    let e = (from_expr [] omega) in *)
(*    let _ = print_string (print e) in *)
(*    let s = print (eval [] e) in *)
(*      print_string s *)

