(* lambda-lifting *)

open Ast

let name_lambdas freshgen expr =
  let rec name freshgen (expr_desc, expr_id) =
    let desc, freshgen = match expr_desc with
      | Eident _
      | Econstant _
      | Econstruct(_, None) ->
	  expr_desc, freshgen
      | Elet(v, bound, body) ->
	  let bound, freshgen = name freshgen bound in
	  let body, freshgen  = name freshgen body in
	  let body, freshgen  = match desc body with
	    | Efunction(_, _)
	    | Efixpoint(_, _, _) ->
		let freshgen, id = Fresh.next freshgen in
		let fv           = Fresh.var id in
		let new_body = mklet fv body (mkident fv) in
		  new_body, freshgen
	    | _ ->
		body, freshgen
	  in
	    Elet(v, bound, body), freshgen
      | Efunction(v, body) ->
	  let body, freshgen = name freshgen body in
	    Efunction(v, body), freshgen
      | Efixpoint(fixv, v, body) ->
	  let body, freshgen = name freshgen body in
	    Efixpoint(fixv, v, body), freshgen
      | Eapply(func, args) ->
	  let func, freshgen = name freshgen func in
	  let args, freshgen = List.fold_right 
	    (fun arg (args, freshgen) ->
	      let arg, freshgen = name freshgen arg in
		match desc arg with
		  | Efunction(_, _)
		  | Efixpoint(_, _, _) ->
		      let freshgen, id = Fresh.next freshgen in
		      let fv           = Fresh.var id in
		      let arg = mklet fv arg (mkident fv) in
			(arg :: args, freshgen)
		  | _ ->
		      (arg :: args), freshgen
	    ) args ([], freshgen) in
	    (match desc func with
	      | Efunction(_, _)
	      | Efixpoint(_, _, _) ->
		  let freshgen, id = Fresh.next freshgen in
		  let fv           = Fresh.var id in
		  let result = mklet fv func (mkexpr (Eapply(mkident fv, args))) in
		    (desc result), freshgen
	      | _ -> 
		  Eapply(func, args), freshgen)
      | Ematch(matched_expr, matching) ->
	  let matched, freshgen  = name freshgen matched_expr in
	  let matching, freshgen = List.fold_right 
	    (fun (pattern, expr) (matching, freshgen) ->
	      let expr, freshgen = name freshgen expr in
		match desc expr with
		  | Efunction(_, _)
		  | Efixpoint(_, _, _) ->
		      let freshgen, id = Fresh.next freshgen in
		      let fv           = Fresh.var id in
		      let match_line = (pattern, mklet fv expr (mkident fv)) in
			(match_line :: matching, freshgen)
		  | _ ->
		      ((pattern, expr) :: matching, freshgen)
	    ) matching ([], freshgen)
	  in
	    (match desc matched with
	      | Efunction(_, _)
	      | Efixpoint(_, _, _) ->
		  let freshgen, id = Fresh.next freshgen in
		  let fv           = Fresh.var id in
		  let result = mklet fv matched (mkmatch (mkident fv) matching) in
		    (desc result), freshgen
	      | _ ->
		  Ematch(matched, matching), freshgen)
      | Etuple fields ->
	  let fields, freshgen = List.fold_right 
	    (fun field (fields, freshgen) ->
	      let field, freshgen = name freshgen field in
		match desc field with
		  | Efunction(_, _)
		  | Efixpoint(_, _, _) ->
		      let freshgen, id = Fresh.next freshgen in
		      let fv           = Fresh.var id in
		      let field = mklet fv field (mkident fv) in
			(field :: fields, freshgen)
		  | _ ->
		      (field :: fields), freshgen
	    ) fields ([], freshgen) in
	    (Etuple fields), freshgen
      | Econstruct(cons, Some e) ->
	  let e, freshgen = name freshgen e in
	    (match desc e with
	      | Efunction(_, _)
	      | Efixpoint(_, _, _) ->
		  let freshgen, id = Fresh.next freshgen in
		  let fv           = Fresh.var id in
		  let e = mklet fv e (mkident fv) in
		    Econstruct(cons, Some e), freshgen
	      | _ ->
		  Econstruct(cons, Some e), freshgen)
    in
      (desc, expr_id), freshgen
  in
    name freshgen expr

let rec box_expr_with_lambdas expr variables =
  match variables with
    | [] ->
	expr
    | x :: l ->
	let res = box_expr_with_lambdas expr l in
	  mkfun x res

let rec apply_free_variables expr variables =
  match variables with
    | [] ->
	expr
    | x :: l ->
	apply_free_variables (mkapp expr (mkident x)) l
	  
let rec lift (expr_desc, expr_id) acc =
  let desc, lifted = match expr_desc with
    | Eident _
    | Econstant _
    | Econstruct(_, None) ->
	expr_desc, acc
    | Elet(var, bound, body) ->
	(match desc bound with
	  | Efunction(_, _)
	  | Efixpoint(_, _, _) ->
	      let bound, lifted = lift bound acc in
	      let body, lifted  = lift body lifted in
	      let fvs = List.map fst (free_variables bound) in
	      let combinator = box_expr_with_lambdas bound fvs in
	      let call_point = apply_free_variables (mkident var) fvs in
	      let body = subst var (desc call_point) body in
	      let lifted = (var, combinator) :: lifted in
		(desc body), lifted
	  | _ ->
	      let bound, lifted = lift bound acc in
	      let body, lifted = lift body lifted in
		Elet(var, bound, body), lifted)
    | Efunction(var, body) ->
	let body, lifted = lift body acc in
	  Efunction(var, body), lifted
    | Efixpoint(vfix, var, body) ->
	let body, lifted = lift body acc in
	  Efixpoint(vfix, var, body), lifted
    | Eapply(func, args) ->
	let func, lifted = lift func acc in
	let args, lifted = List.fold_right (fun arg (args, lifted) ->
	  let arg, lifted = lift arg lifted in
	    (arg :: args, lifted)
	) args ([], lifted) in
	  Eapply(func, args), lifted
    | Ematch(matched_expr, matching) ->
	let matched, lifted  = lift matched_expr acc in
	let matching, lifted = List.fold_right
	  (fun (pattern, expr) (matching, lifted) ->
	    let expr, lifted = lift expr lifted in
	      ((pattern, expr) :: matching, lifted)
	  ) matching ([], lifted) in
	  Ematch(matched, matching), lifted
    | Etuple fields ->
	let fields, lifted = List.fold_right
	  (fun field (fields, lifted) ->
	    let field, lifted = lift field lifted in
	      (field :: fields, lifted)
	  ) fields ([], acc) in
	  (Etuple fields), lifted
    | Econstruct(cons, Some e) ->
	let e, lifted = lift e acc in
	  (Econstruct(cons, Some e)), lifted
  in
    (desc, expr_id), lifted

let lambda_lift (decls, main) freshgen =
  let named, freshgen = name_lambdas freshgen main in 
  let main, lifted = lift named [] in
  let prog = List.fold_left (fun expr (binder, combinator) ->
    mklet binder combinator expr
  ) main lifted in
    freshgen, (decls, prog)
