(*s This module handles pattern-matching compilation. The aim is to
  transform complex pattern-matchings with possibly deeply nested
  patterns into a simple tree of tests on constructors. *)

(*i*)
open Ast
(*i*)

(*s The pattern matching algorithm works on a matrix of patterns.
  We use a very inefficient but handy representation of matrices: 
  lists of lists. *)
type 'a matrix = 'a line list
and  'a line   = 'a list

(*s We will need a way to distinguish between the various patterns.
  This type is used at this end. *)
type 'a constructor_kind =
  | Algebraic of string * 'a
  | Constant of Ast.constant

(*s The pattern matching algorithm generates a [tree]. There
  is no sharing of identical sub-trees. *)
type tree =
  | UnfoldTuple     of Ast.vvar * Ast.vvar list * tree
  | SwitchConstruct of Ast.vvar * datacons_case list * default_case
  | SwitchConstant  of Ast.vvar * constant_case list * default_case
  | Atom of expr
and datacons_case = Ast.datacons * Ast.vvar option * tree
and constant_case = Ast.constant * tree
and default_case = tree option

(*s We will need various tools to manipulate matrices. *)

(*s [slice_line] slices a line [l] into $[l']|[x]|[l'']$ where [x] is the n-th element. *)
let slice_line l n =
  let rec aux n acc l =
    match l with
      | [] ->
	  failwith "slice_line : line too short."
      | x :: l'' ->
	  if n = 1 then
	    (List.rev acc, x, l'')
	  else
	    aux (n-1) (x :: acc) l''
  in aux n [] l

(* Slice a matrix m into m'|c|m'' where c is the n-th column. *)
let slice_column m n =
  let sliced_lines = List.map (fun l -> slice_line l n) m in
  let m', c, m''    = List.fold_right (fun (l', x, l'') (m', c, m'') ->
    (l' :: m', x :: c, l'' :: m'')
  ) sliced_lines ([], [], []) in
    (m', c, m'')

(* Merges a left list l', a middle _list_ x and a right list l'' into a single list *)
let unslice_line l' x l'' = l' @ (x @ l'')

(* Merges a left matrix m', a middle _matrix_ c and a right matrix m'' into a new matrix *)
let unslice_column m' c m'' =
  let sliced_lines = List.combine (List.combine m' c) m'' in
    List.fold_right (fun ((l', x), l'') m ->
      (unslice_line l' x l'') :: m
    ) sliced_lines []

let rec remove_at idx l =
  if idx = 1 then
    List.tl l
  else match l with
    | [] ->
	failwith "remove_at : out of list bounds"
    | x :: l' ->
	x :: (remove_at (idx - 1) l')

let get_column matrix column =
  let idx = column - 1 in (* List.nth counts from 0 *)
    List.map (fun (line, _) -> List.nth line idx) matrix

(*i ------------------------------------------------------------ i*)
(*i Various functions useful in the match compilation algorithm. i*)

let rec pattern_matches_all (patt_desc, _) =
  match patt_desc with
    | Pany -> true
    | Pvar _ -> true
    | Ptuple patts -> false
	(* List.for_all pattern_matches_all patts *)
    | _ -> false

(* Returns true iff a line or a column matches everything. *)
let rec list_matches_all list =
  List.for_all pattern_matches_all list

(* Returns true iff a given column of a matrix matches everything. *)
let column_matches_all matrix column =
  list_matches_all (get_column matrix column)

(* Makes a list of [length] "Pany", i.e. a line of [length] catch-alls *)
let rec make_catch_all length =
  if length = 0 then
    []
  else
    Ast.mkpany :: (make_catch_all (length - 1))

let make_fresh_variables freshgen length =
  let rec mkvars freshgen variables length =
    if length = 0 then
      freshgen, variables
    else
      let gen, fresh = Fresh.next freshgen in
      let v'         = Fresh.var fresh in
	mkvars gen (v' :: variables) (length - 1)
  in
    mkvars freshgen [] length

(* Makes a list of [length] fresh variables. *)
let make_fresh_variables freshgen length =
  let rec mkvars freshgen variables length =
    if length = 0 then
      freshgen, variables
    else
      let gen, fresh = Fresh.next freshgen in
      let v'         = Fresh.var fresh in
	mkvars gen (v' :: variables) (length - 1)
  in
    mkvars freshgen [] length

let rec test_count tree =
  match tree with
    | UnfoldTuple(_, _, tr) ->
	(test_count tr)
    | SwitchConstruct(_, cases, default) ->
	let res = List.fold_left (fun count (_, _, tr) ->
	  (test_count tr) + count
	) 1 cases in
	  (match default with
	    | None -> res
	    | Some tr ->
		res + (test_count tr))
    | SwitchConstant(_, cases, default) ->
	let res = List.fold_left (fun count (_, tr) ->
	  (test_count tr) + count
	) 1 cases in
	  (match default with
	    | None -> res
	    | Some tr ->
		res + (test_count tr))
    | Atom _ -> 0

let cost_measure = test_count

let backtrack_minimize : (int -> 'a option) -> ('a -> int) -> int -> int -> 'a option =
  fun compute cost_measure left right ->
    let rec search index current_min min_cost =
      if index > right then
	current_min
      else
	match compute index with
	  | None ->
	      search (index + 1) current_min min_cost
	  | Some res ->
	      let cost = cost_measure res in
		if cost < min_cost then
		  search (index+1) (Some res) cost
		else
		  search (index+1) current_min min_cost
    in search left None max_int

(* The match compilation algorithm is functorized. *)

module type InputSig =
sig

  (* [sig_of x] returns the list of all constructors defined along [x],
   * including [x]. *)
  val sig_of : Ast.datacons -> Ast.datacons list

end


module Make(I : InputSig)
  =
struct

  (* Simplify a matrix knowing that on [column] of [matrix],
   * the matched term is [constant].  *)
  let simplify_constant matched_variable constant column matrix =
    List.fold_right (fun (line, action) new_matrix ->
      let l', (x, _), l'' = slice_line line column in
	match x with
	  | Pany ->
	      let line = remove_at column line in
		(line, action) :: new_matrix
	  | Pvar v ->
	      let action = subst v (Eident matched_variable) action in
	      let line = remove_at column line in
		(line, action) :: new_matrix
	  | Pconst c ->
	      if c = constant then
		let line = remove_at column line in
		  (line, action) :: new_matrix
	      else
		new_matrix
	  | Pconstruct(_, _) ->
	      new_matrix
	  | Ptuple _ ->
	      failwith "badly flattened matrix"
    ) matrix []


  (* Simplify a matrix knowing that on [column] of [matrix],
   * the matched term is of shape [constructor](..).
   * [arity] is either 0 or 1. *)
  let simplify_algebraic matched_variable constructor arity column matrix =
    List.fold_right (fun (line, action) new_matrix ->
      let l', (x, _), l'' = slice_line line column in
	match x with
	  | Pany ->
	      let catch_all = make_catch_all arity in
	      let line = unslice_line l' catch_all l'' in
		(line, action) :: new_matrix
	  | Pvar v ->	      
	      let action = subst v (Eident matched_variable) action in
	      let catch_all = make_catch_all arity in
	      let line = unslice_line l' catch_all l'' in
		(line, action) :: new_matrix
	  | Pconst _ ->
	      new_matrix
	  | Pconstruct(cons, None) ->
	      if cons = constructor then
		let line = remove_at column line in
		  (line, action) :: new_matrix
	      else
		new_matrix
	  | Pconstruct(cons, Some field) ->
	      if cons = constructor then
		let line = unslice_line l' [field] l'' in
		  (line, action) :: new_matrix
	      else
		new_matrix
	  | Ptuple _ ->
	      failwith "badly flattened matrix"
    ) matrix []


  (* Simplify a matrix keeping all the lines where the pattern
     on [column] is a catch-all or a variable. *)
  let default matched_variable column matrix =
    List.fold_right (fun (line, action) new_matrix ->
      let l', (x, _), l'' = slice_line line column in
	match x with
	  | Pany ->
	      let line = remove_at column line in
		(line, action) :: new_matrix
	  | Pvar v ->
	      let action = subst v (Eident matched_variable) action in
	      let line = remove_at column line in
		(line, action) :: new_matrix
	  | Pconst _ 
	  | Pconstruct(_, _) ->
	      new_matrix
	  | Ptuple _ ->
	      failwith "badly flattened matrix"
    ) matrix []


  let collect_constructors column =
    List.fold_left (fun acc (patt_desc, _) ->
      match patt_desc with
	| Pany | Pvar _ -> acc
	| Pconst c ->
	    if List.mem (Constant c) acc then
	      acc
	    else
	      (Constant c) :: acc
	| Ptuple _ ->
	    failwith "badly flattened matrix"
	| Pconstruct(cons, field) ->
	    let arity =
	      if field = None then 0 else 1
	    in
	      if List.mem (Algebraic(cons, arity)) acc then
		acc
	      else
		(Algebraic(cons, arity)) :: acc
    ) [] column

  (* Pattern-matching compilation algorithm. *)
  let rec compile vars matrix freshgen =
    let first_line, first_action = match matrix with
      | [] ->
	  failwith "Empty pattern matrix."
      | (line, act) :: _ -> 
	  (line, act)
    in
      if list_matches_all first_line then
	(* The first line matches everything. The first action
	   will thus be executed. Matching compilation ends. *)
	let action = List.fold_left (fun action (var, (pattern_desc, _)) ->
	  match pattern_desc with
	    | Pany -> action
	    | Pvar v ->
		subst v (Eident var) action
	    | _ ->
		failwith "bug in Matching"
	) first_action (List.combine vars first_line) 
	in
	  freshgen, (Atom action)
      else
	(*backtrack_minimize f cost 0 width*)
	let explore column_idx =
	  match List.nth first_line (column_idx-1) with
	    | (Ptuple fields, _) ->
		Some (compile_tuple vars matrix fields column_idx freshgen)
	    | _ ->
		(* Choose a column which is not all-wilcard *)
		compile_target_column vars matrix freshgen column_idx
	in
	let cost (_, tree) = cost_measure tree in
	  match backtrack_minimize explore cost 1 (List.length first_line) with
	    | None ->
		failwith "no solution found while compiling matching"
	    | Some res ->
		res
		  
  and compile_target_column vars matrix freshgen column_idx =
    (* Not a tuple pattern column. 
     * Get all the constructors used in the column. *)
    let column = get_column matrix column_idx in
      if list_matches_all column then
	(* This column is not valid. *)
	None
      else
	let constructors = collect_constructors column in
	  (* Build the switch for the constructors *)
	let result = match constructors with
	  | [] -> failwith "no constructors in column"
	  | (Constant _) :: _ ->
	      (* We match on constants. *)
	      let constructors = List.map (function Constant const -> const
		| _ -> failwith "non-constant case in constant match"
	      ) constructors
	      in
		compile_constant vars matrix constructors column_idx freshgen
	  | _ ->
	      (* We match on an algebraic datatype *)
	      let constructors = List.map (function 
		| Algebraic(cons, 0) -> (cons, 0)
		| Algebraic(cons, 1) -> (cons, 1)
		| _ -> failwith "non-algebraic case in constant match"
	      ) constructors 
	      in
		compile_algebraic vars matrix constructors column_idx freshgen
	in
	  Some result

  and compile_tuple vars matrix fields column_idx freshgen =
    (* If the chosen column is a tuple pattern, just unfold 
       the pattern using a dummy match. *)
    let arity      = List.length fields in
    let gen, fvs   = make_fresh_variables freshgen arity in
    let l', x, l'' = slice_line vars column_idx in
    let vars       = unslice_line l' fvs l'' in
    let mat, acts  = List.split matrix in
    let m', c, m'' = slice_column mat column_idx in
    let new_column = List.map (fun patt ->
      match patt with
	| (Ptuple fields, _) -> 
	    fields
	| (Pany, _) ->
	    make_catch_all arity
	| _ -> failwith "bug in matching compilation."
    ) c in
    let columns   = unslice_column m' new_column m'' in
    let matrix    = List.combine columns acts in
    let gen, code = compile vars matrix gen in
    let code      = UnfoldTuple(x, fvs, code) in
      gen, code

  and compile_constant vars matrix constants column_idx freshgen =
    let l', matched_variable, l'' = slice_line vars column_idx in
    let vars = unslice_line l' [] l'' in
    let default_matrix = default matched_variable column_idx matrix in
    let freshgen, default_code = match default_matrix with
	| [] -> (* Empty default matrix, no default branch *)
	    freshgen, None
	| _ ->
	    let freshgen, result = compile vars default_matrix freshgen in
	      freshgen, (Some result)
    in
    let freshgen, cases = 
      List.fold_left (fun (freshgen, cases) const ->
	let matrix    = simplify_constant matched_variable const column_idx matrix in
	let gen, code = compile vars matrix freshgen in
	  (gen, (const, code) :: cases)
      ) (freshgen, []) constants
    in
    let code = SwitchConstant(matched_variable, cases, default_code) in
      freshgen, code

  and compile_algebraic vars matrix constructors column_idx freshgen =
    let l', matched_variable, l'' = slice_line vars column_idx in
    let vars = unslice_line l' [] l'' in
    let default_matrix = default matched_variable column_idx matrix in
    let freshgen, default_code = match default_matrix with
	| [] -> (* Empty default matrix, no default branch *)
	    freshgen, None
	| _ ->
	    let freshgen, result = compile vars default_matrix freshgen in
	      freshgen, (Some result)
    in
    let freshgen, cases = 
      List.fold_left (fun (freshgen, cases) (cons, arity) ->
	if arity = 0 then
	  let matrix    = simplify_algebraic matched_variable cons 0 column_idx matrix in
	  let gen, code = compile vars matrix freshgen in
	    (gen, (cons, None, code) :: cases)
	else
	  (* arity = 1 *)
	  let matrix     = simplify_algebraic matched_variable cons 1 column_idx matrix in
	  let gen, fresh = Fresh.next freshgen in
	  let v'         = Fresh.var fresh in
	  let vars       = unslice_line l' [v'] l'' in
	  let gen, code  = compile vars matrix gen in
	    (gen, (cons, (Some v'), code) :: cases)
      ) (freshgen, []) constructors
    in
    let code = SwitchConstruct(matched_variable, cases, default_code) in
      freshgen, code


  (*s After pattern matching compilation, we use [tree_to_expr] to convert a
    test tree back into a now simple match construct. *)
  let rec tree_to_expr tree =
    match tree with
      | UnfoldTuple(var, tuplevars, subnode) ->
	  let pattvars = List.map Ast.mkpvar tuplevars in
	  let patts    = mkptuple pattvars in
	    mkmatch (mkident var)
	      [(patts, (tree_to_expr subnode))]
      | SwitchConstruct(var, cases, default) ->
	  let cases = List.map (fun (datacons, content, action) ->
	    (match content with
	      | None ->
		  (mkpconstruct datacons None, tree_to_expr action)
	      | Some v ->
		  (mkpconstruct datacons (Some (mkpvar v)), tree_to_expr action))
	  ) cases in
	  let cases = match default with
	    | None -> cases
	    | Some code ->
		cases @ [(mkpany, tree_to_expr code)]
	  in
	    mkmatch (mkident var) cases
      | SwitchConstant(var, cases, default) ->
	  let cases = List.map (fun (constant, action) ->
	    (mkpconst constant, tree_to_expr action)
	  ) cases in
	  let cases = match default with
	    | None -> cases
	    | Some code ->
		cases @ [(mkpany, tree_to_expr code)]
	  in
	    mkmatch (mkident var) cases
      | Atom e -> e

  let rec matching_to_matrix matching =
    List.map (fun (pattern, action) -> ([pattern], action)) matching

  let rec compile_matching matched_expr matching freshgen =
    let (matched_desc, matched_id) = matched_expr in
      match matched_desc with
	| Eident id ->
	    let matrix = matching_to_matrix matching in
	    let freshgen, tree = compile [id] matrix freshgen in
	    let matching, _ = tree_to_expr tree in
	      freshgen, matching
	| _ ->
	    let gen, fresh = Fresh.next freshgen in
	    let v'         = Fresh.var fresh in
	    let matrix     = matching_to_matrix matching in
	    let freshgen, tree = compile [v'] matrix gen in
	    let matching   = tree_to_expr tree in
	    let result, _  = mklet v' matched_expr (matching) in
	      freshgen, result

  let rec compile_expr expr freshgen =
    let (expr_desc, expr_id) = expr in
    let freshgen, desc = match expr_desc with
      | Eident _
      | Econstant _
      | Econstruct(_, None) ->
	  freshgen, expr_desc
      | Elet(var, bound, body) ->
	  let freshgen, bound = compile_expr bound freshgen in
	  let freshgen, body  = compile_expr body freshgen in
	    freshgen, Elet(var, bound, body)
      | Efunction(var, body) ->
	  let freshgen, body  = compile_expr body freshgen in
	    freshgen, Efunction(var, body)
      | Efixpoint(vfix, vvar, body) ->
	  let freshgen, body  = compile_expr body freshgen in
	    freshgen, Efixpoint(vfix, vvar, body)
      | Eapply(func, args) ->
	  let freshgen, func = compile_expr func freshgen in
	  let freshgen, args = List.fold_right (fun arg (freshgen, args) ->
	    let freshgen, arg = compile_expr arg freshgen in
	      (freshgen, arg :: args)
	  ) args (freshgen, []) in
	    freshgen, (Eapply(func, args))
      | Ematch(matched_expr, matching) ->
	  let freshgen, matched_expr = compile_expr matched_expr freshgen in
	  let freshgen, matching = List.fold_right 
	    (fun (pattern, action) (freshgen, matching) ->
	      let freshgen, action = compile_expr action freshgen in
		(freshgen, (pattern, action) :: matching)
	    ) matching (freshgen, [])
	  in
	    compile_matching matched_expr matching freshgen
      | Etuple fields ->
	  let freshgen, fields = List.fold_right 
	    (fun field (freshgen, fields) ->
	      let freshgen, field = compile_expr field freshgen in
		(freshgen, field :: fields)
	    ) fields (freshgen, []) in
	    freshgen, (Etuple fields)
      | Econstruct(datacons, Some e) ->
	  let freshgen, e = compile_expr e freshgen in
	    freshgen, (Econstruct(datacons, Some e))
    in
      freshgen, (desc, expr_id)

  let rec compile_program (decls, main) freshgen =
    let freshgen, main = compile_expr main freshgen in
      freshgen, (decls, main)


end
  
