(* Types *)

(*s  This modules defines our types, wich are in fact simple 
  first-order terms with some distinguished constructors
  (tuple, arrow \ldots). *)

(*i*)
open Printf
(*i*)

open Terms

(* ------------------------------------------------------------- *)

(* Some operations on type & type variables *)

(* Peforms a 'fold' on the type structure. *)
let rec fold fv fc =
  function
    | TermVar x -> fv x
    | TermCons(cons, subtrees) ->
	let sub' = List.map (fold fv fc) subtrees in
	  fc cons sub'
	    
(* Returns a list of all free type variables in [typ].
   (Since there is no explicit quantification, all variables are 'free' ...) *)
let free_type_variables typ =
  fold 
    (fun v -> [v])
    (fun _ subterms -> List.flatten subterms)
    typ

(* A type is ground if it contains no variables. *)
let rec is_ground =
  function
    | TermCons(_, subterms) ->
	List.for_all is_ground subterms
    | TermVar _ -> false

(* Instantiates a type variable for a (possibly non-ground) type in [typ] *)
let rec inst_type_variable typ v new_typ =
  match typ with
    | TermCons(x, subterms) ->
	let subterms = List.map (fun subtyp -> 
	  inst_type_variable subtyp v new_typ) subterms
	in
	  TermCons(x, subterms)
    | TermVar v' ->
	if v = v' then
	  new_typ
	else
	  typ 

let label_of_type =
  function
    | TermVar _ -> None
    | TermCons(x, _) -> Some x

(* ------------------------------------------------------------- *)

module type TypeSig =
sig
  type term
  val types_equivalent : term -> term -> bool
  val print : term -> string
end


(* The two following modules, Default and NoDebug, are implementations of
   the TypeSig ande unification.TermSig signatures. Default.term types
   are annotated with some debug informations inside the constructor label.
   NoDebug terms are bare-bones. *)

module Default =
struct

  type debug_info        = string

  type constructor_label = debug_info * Ast.typecons

  type term              = constructor_label Unification.pterm


  let label_compatible : constructor_label -> constructor_label -> bool =
    fun (_, s1) (_, s2) ->
      s1 = s2

  let string_of_label : constructor_label -> string =
    fun (debug, label) -> label

  let debug_of_label : constructor_label -> string =
    fun (debug, label) -> debug


  (* alpha-equivalence *)
  let types_equivalent t1 t2 =
    let rec alpha_convert_type mapping seed =
      function
	| TermVar v ->
	    if List.mem_assoc v mapping then
	      (TermVar (List.assoc v mapping), mapping, seed)
	    else
	      let next = seed+1 in
		(TermVar seed, (v, next) :: mapping, next)
	| TermCons(x, subterms) ->
	    let subterms, mapping, seed = List.fold_left (fun (subterms, mapping, seed) term ->
	      let term, mapping, seed = alpha_convert_type mapping seed term in
		(term :: subterms, mapping, seed)
	    ) ([], mapping, seed) subterms in
	      (TermCons(x, List.rev subterms), mapping, seed)
    in
    let rec compare t1 t2 =
      match (t1, t2) with
	| TermVar v1, TermVar v2 -> v1 = v2
	| TermCons(x1, subterms1), TermCons(x2, subterms2) ->
	    if List.length subterms1 = List.length subterms2 then
	      (label_compatible x1 x2) && (List.for_all2 compare subterms1 subterms2)
	    else
	      false
	| _ -> false
    in
    let (t1, _, _) = alpha_convert_type [] 0 t1 in
    let (t2, _, _) = alpha_convert_type [] 0 t2 in
      compare t1 t2

  (* change the debug info within a type *)
  let set_debug debug_info =
    function
      | TermVar v -> TermVar v
      | TermCons((_, label), subterms) ->
	  TermCons((debug_info, label), subterms)
	

  (* ugly-printing functions *)

  let print x =
    Terms.string_of_term
      string_of_int
      string_of_label
      x

  (* Some helpful functions : combinators to create types. *)

  let mkvar v =
    TermVar v

  let mktype debug head subterms =
    TermCons((debug, head), subterms)

  let mk0ary debug head =
    TermCons((debug, head), [])

  let arrow_type debug dom codom =
    mktype debug "->" [ dom; codom ]

  let tuple_type debug  l =
    mktype debug "*" l

  let array_type debug t =
    mktype debug "array" [ t ]

  let string_type debug =
    mktype debug "string" []

  let float_type debug =
    mktype debug "float" []

  let int_type debug =
    mktype debug "int" []

  let int32_type debug =
    mktype debug "int32" []

  let int64_type debug =
    mktype debug "int64" []

  let nativeint_type debug =
    mktype debug "nativeint" []

  let char_type debug =
    mktype debug "char" []

  let bool_type debug =
    mktype "bool" []
      
  let unit_type debug =
    mktype debug "unit" []


end

module NoDebug =
struct

  type constructor_label = Ast.typecons

  type term              = constructor_label Unification.pterm

  let types_equivalent t1 t2 =
    let rec alpha_convert_type mapping seed =
      function
	| TermVar v ->
	    if List.mem_assoc v mapping then
	      (TermVar (List.assoc v mapping), mapping, seed)
	    else
	      let next = seed+1 in
		(TermVar seed, (v, next) :: mapping, next)
	| TermCons(x, subterms) ->
	    let subterms, mapping, seed = List.fold_left (fun (subterms, mapping, seed) term ->
	      let term, mapping, seed = alpha_convert_type mapping seed term in
		(term :: subterms, mapping, seed)
	    ) ([], mapping, seed) subterms in
	      (TermCons(x, List.rev subterms), mapping, seed)
    in
    let (t1, _, _) = alpha_convert_type [] 0 t1 in
    let (t2, _, _) = alpha_convert_type [] 0 t2 in
      t1 = t2

  (* ugly-printing functions *)

  let print x =
    Terms.string_of_term
      string_of_int
      (fun x -> x)
      x

  (* Some helpful functions : combinators to create types. *)

  let mkvar v =
    TermVar v

  let mktype head subterms =
    TermCons(head, subterms)

  let mk0ary head =
    TermCons(head, [])

  let arrow_type dom codom =
    mktype "->" [ dom; codom ]

  let tuple_type  l =
    mktype "*" l

  let array_type t =
    mktype "array" [ t ]

  let string_type =
    mktype "string" []

  let float_type =
    mktype "float" []

  let int_type =
    mktype "int" []

  let int32_type =
    mktype "int32" []

  let int64_type =
    mktype "int64" []

  let nativeint_type =
    mktype "nativeint" []

  let char_type =
    mktype "char" []

  let bool_type =
    mktype "bool" []
      
  let unit_type =
    mktype "unit" []

end


(* We need a function to convert from Default to NoDebug terms. *)
let forget_debug x = 
  Terms.term_lift
    (fun (debug, cons) -> cons)
    (fun var -> Terms.TermVar var) x


module AlphaTypeMap(I : TypeSig) =
struct

  type 'a t = (I.term * 'a) list

  let empty = []

  let rec mem key map =
    match map with
      | [] -> false
      | (t, _) :: tl ->
	  if I.types_equivalent key t then
	    true
	  else
	    mem key tl

  let rec find_opt key map =
    match map with
      | [] -> None
      | (t, elt) :: tl ->
	  if I.types_equivalent key t then
	    Some elt
	  else
	    find_opt key tl
	      
  let add key elt map =
    (key, elt) :: (List.filter (fun (key', _) -> key' <> key) map)

  let rec fold f map acc =
    match map with
      | [] -> acc
      | (x, y) :: l ->
	  fold f l (f x y acc)

  let fusion op map1 map2 =
    fold (fun domain codomain map ->
      match find_opt domain map with
	| None ->
	    add domain codomain map
	| Some elt ->
	    add domain (op elt codomain) map
    ) map1 map2

  let domain map = List.map fst map

  let codomain map = List.map snd map

end


(* Type set quotiented by alpha-equivalence. *)
module AlphaTypeSet(I : TypeSig) = 
struct
  include Set.Make(
    struct
      type t = I.term
	  
      let compare t1 t2 = 
	if I.types_equivalent t1 t2 then
	  0 (* alpha-equivalent types *)
	else
	  Pervasives.compare t1 t2
    end)

  let from_set (s : I.term Pset.t) =
    Pset.fold add s empty

end

